{rem$define lpmapping}
unit Matrix;
{
Operations with matrix contains fractional numbers
for Delphi 3.0
Copyright (c) 1996-1997 Dmitriy Anisimkov
e-mail anisimkov@yahoo.com
phone +7 (381)-2217774
}
interface

uses MxEl //this is a alias for unit Rational
{$ifdef lpmapping}
,windows
{$endif}
;

type

pElement = ^tElement;

tRangeInMem = 0..MAXINT div sizeof(tElement)-1;
pArrayOfElements = ^tArrayOfElements;
tArrayOfElements = array [tRangeInMem] of tElement;

tMatrix =class
private
  fMaxI,fMaxJ,fRowsCount,fColsCount : integer;
  MemSize : integer;
{$ifdef lpmapping}
  MapFileName : String;
  HMap,HFile : THandle;
{$endif}
  procedure InitSize(Rows,Cols : integer);
protected
  Datas : pArrayOfElements;
  procedure SetMaxI(mi : integer);
  procedure SetMaxJ(mj : integer);
  procedure SetCols(c : integer);
  procedure SetRows(r : integer);
  function GetElement(i,j : integer) : tElement;
  procedure SetElement(i,j : integer;const it : tElement);
  function Idx(i,j : integer) : integer;
  procedure AddRow;
  procedure Jordan(r,s : integer);
  procedure SumRow(r1,r2 : integer);
  procedure ExchangeRows(i1,i2 : integer); virtual;
  procedure ExchangeColumns(j1,j2 : integer);
public
  constructor Create(Rows,Cols : integer);
  constructor Copy(from : TMatrix); virtual;
  constructor PartCopy(from : TMatrix;bi,bj,Rows,Cols : integer);
  constructor Generate(Size : integer);
  constructor Generate2(Size : integer);
  constructor Generate3(Size : integer);
  destructor Destroy; override;
  property a[i,j : integer] : tElement read GetElement write SetElement; default;
  property RowCount : integer read fRowsCount;
  property ColCount : integer read fColsCount;
  property MaxI : integer read FMaxI;
  property MaxJ : integer read FMaxJ;
  procedure SelfInverse;
  function Multiply(m : TMatrix) : TMatrix;
end;

implementation

uses sysutils
,utils
;

type
pElementRec = ^tElementRec;
tElementRec = record
  e : tElement;
end;

procedure Error(msg : String);
begin
  raise Exception.Create(Msg);
end;

procedure tMatrix.InitSize(Rows,Cols : integer);
{$ifdef lpmapping}
var FN : array [0..MAX_PATH] of char;
{$endif}
begin
  SetRows(Rows);
  SetCols(Cols);
  {$Q+}
  MemSize := RowCount*ColCount*sizeof(tElement);
{$ifdef lpmapping}
  ChkBool(GetTempFileName(pChar(GetCurrentDir),'LPM',0,FN)<>0);
  MapFileName := FN;
  HFile := CheckHandle(FileCreate(MapFileName));
  HMap := CreateFileMapping({HFile}-1,nil,PAGE_READWRITE,0,MemSize,'');
  ChkBool(HMap>0);
  Datas := MapViewOfFile(HMap,FILE_MAP_ALL_ACCESS,0,0,MemSize);
  ChkBool(Assigned(Datas));
{$else}
  Datas := AllocMem(MemSize);
{$endif}
  {$Q-}
end;

procedure tMatrix.SetMaxI(mi : integer);
begin
  FMaxI := mi;
  FRowsCount := mi+1;
end;
procedure tMatrix.SetMaxJ(mj : integer);
begin
  FMaxJ := mj;
  FColsCount := mj+1;
end;
procedure tMatrix.SetRows(r : integer);
begin
  FMaxI := r-1;
  FRowsCount := r;
end;
procedure tMatrix.SetCols(c : integer);
begin
  FMaxJ := c-1;
  FColsCount := c;
end;

constructor tMatrix.Create(Rows,Cols : integer);
begin
  InitSize(Rows,Cols);
  for Rows := 0 to RowCount*ColCount-1 do Datas^[Rows] := NULL;
end;

destructor tMatrix.Destroy;
begin
{$ifdef lpmapping}
   ChkBool(UnmapViewOfFile(Datas));
   ChkBool(CloseHandle(HMap));
   ChkBool(CloseHandle(HFile));
   DeleteFile(MapFileName);
{$else}
   FreeMem(Datas);
{$endif}
  inherited Destroy;
end;

constructor tMatrix.Generate(Size : integer);
var
  i,j : integer;
begin
  InitSize(size,size);
  for i := 0 to MaxI do for j := 0 to MaxJ do begin
    with Datas[idx(i,j)] do begin
       numer := i+j+1;
       if numer>ColCount then numer := numer-ColCount;
       denom := 2*ColCount-i-j-1;
    end;
  end;
end;

constructor tMatrix.Generate3(Size : integer);
var
  i,j : integer;
begin
  InitSize(size,size);
  for i := 0 to MaxI do for j := 0 to MaxJ do begin
    with Datas[idx(i,j)] do begin
       numer := i+j+1;
       if numer>ColCount then numer := numer-ColCount;
       denom := 1;
       if (MaxI-i>j) then ExchangeMem(numer,denom,sizeof(numer));
    end;
  end;
end;

constructor tMatrix.Generate2(Size : integer);
var
  i,j : integer;
begin
  InitSize(size,size);
  for i := 0 to MaxI do for j := 0 to MaxJ do begin
    Datas[idx(i,j)] := ONE;
    if i<=j then Datas[idx(i,j)].numer := 2;
  end;
end;

constructor tMatrix.Copy(from : TMatrix);
begin
  InitSize(From.RowCount,From.ColCount);
  Move(from.Datas^,Datas^,MemSize);
end;

constructor tMatrix.PartCopy(from : TMatrix;bi,bj,Rows,Cols : integer);
var
  i,j : integer;
begin
  InitSize(Rows,Cols);
  for i := 0 to MaxI do for j := 0 to MaxJ do Datas[idx(i,j)] :=
    From.Datas^[From.idx(i+bi,j+bj)];
end;

procedure tMatrix.SumRow(r1,r2 : integer);
var
  s : integer;
begin
  for s:=0 to MaxJ do with tElementRec(Datas[idx(r1,s)]) do
    e := Sum(e,Datas[idx(r2,s)]);
end;

function tMatrix.Multiply(m : TMatrix) : TMatrix;
var
  i,j,cn : integer;
begin
  result := TMatrix.Create(RowCount,m.ColCount);
  for i := 0 to result.MaxI do for j := 0 to result.MaxJ do
  with TElementRec(result.Datas[result.idx(i,j)]) do
  for cn := 0 to MaxJ do
    e := Sum(e,mxel.Multiply(Datas[idx(i,cn)],m.Datas[m.idx(cn,j)]));
end;

procedure tMatrix.ExchangeRows(i1,i2 : integer);
var
  j : integer;
begin
  if i1=i2 then exit;
  for j := 0 to MaxJ do
    ExchangeMem(Datas[idx(i1,j)],Datas[idx(i2,j)],sizeof(Datas[0]));
end;

procedure tMatrix.ExchangeColumns(j1,j2 : integer);
var
  i : integer;
begin
  if j1=j2 then exit;
  for i := 0 to MaxI do
    ExchangeMem(Datas[idx(i,j1)],Datas[idx(i,j2)],sizeof(Datas[0]));
end;

procedure tMatrix.SelfInverse;
var
  i,j : integer;
  SetOfVar : ^tArrayOfInteger;
label break2;
begin
  if RowCount<>ColCount then Error('Unposable');
  SetOfVar := AllocMem(RowCount*sizeof(SetOfVar[0]));
  try
  for i := 0 to MaxI do begin
    for j := i to MaxJ do if Compare(Datas[idx(i,j)],NULL)<>0 then begin
      Jordan(i,j);
      ExchangeColumns(i,j);
      SetOfVar[i] := j;
      goto Break2;
    end;
    Error('Range of matrix lower then size');
break2:
  end;
  for i := 0 to MaxI do for j := 0 to MaxJ do
  if SetOfVar[j]=i then begin
    ExchangeRows(i,j);
    ExchangeMem(SetOfVar[j],SetOfVar[i],sizeof(SetOfVar[0]));
  end;
  finally
    FreeMem(SetOfVar);
  end;
end;

procedure tMatrix.Jordan(r,s : integer);
var
  i,j : integer;
  NegInvArs,InvArs,NegArjDivArs : tElement;
  Aij,
  Ais : pArrayOfElements;
begin
  InvArs := MxEl.Inverse(A[r,s]);
  NegInvArs := Negative(InvArs);
  Ais := pArrayOfElements(@Datas[idx(0,s)]);
  for j := 0 to MaxJ do if j<>s then begin
    aij := pArrayOfElements(@Datas[idx(0,j)]);
    NegArjDivArs := mxel.Multiply(Datas^[idx(r,j)],NegInvArs);
    for i := 0 to MaxI do if i=r
    then Aij[i] := mxel.Multiply(Aij[i],InvArs)
    else aij[i] := Sum(aij[i],mxel.Multiply(Ais[i],NegArjDivArs));
  end;
  for i:=0  to MaxI do with tElementRec(Ais[i]) do
    e := mxel.Multiply(e,NegInvArs);
  A[r,s] := InvArs;
end;

function TMatrix.Idx(i,j : integer) : integer;
begin
  result := i{$ifdef ByRows}*ColCount+
  {$else}+RowCount*{$endif}j;
end;

function TMatrix.GetElement(i,j : integer) : tElement;
begin
  result := Datas[idx(i,j)];
end;

procedure TMatrix.SetElement(i,j : integer;const it : tElement);
begin
  Datas[idx(i,j)] := it;
end;

procedure TMatrix.AddRow;
var
  size : integer;
  A0j : pArrayOfElements;
  j : integer;
begin
  size := (RowCount+1)*ColCount*sizeof(tElement);
  if size>MemSize then begin
    ReallocMem(Datas,size);
    MemSize := size;
  end;
  A0j := pArrayOfElements(@Datas[idx(0,MaxJ)]);
  for j := MaxJ downto 1 do begin
    move(A0j[0],A0j[J],RowCount*sizeof(tElement));
    dec(integer(A0j),RowCount*sizeof(tElement));
  end;
  SetRows(RowCount+1);
end;

end.
